from typing import Literal


def select_lang(lang_tag: Literal['en', 'zh'] | str = 'en'):
    if lang_tag == 'en':
        from . import en
        return en
    elif lang_tag == 'zh':
        from . import zh
        return zh
    raise ValueError(f'Invalid lang: {lang_tag}')



import os

from langchain.chat_models import init_chat_model
from langchain.embeddings import init_embeddings


def new_model(
        model_prefix_key: str,
):
    model_type = os.getenv(f"{model_prefix_key}_MODEL_TYPE", default='chat')
    model_provider = os.getenv(f"{model_prefix_key}_MODEL_PROVIDER", default='openai')
    if model_type == 'chat':
        return init_chat_model(
            model=os.getenv(f"{model_prefix_key}_MODEL"),
            model_provider=model_provider,
            base_url=os.getenv(f"{model_prefix_key}_BASE_URL"),
            openai_api_key=os.getenv(f"{model_prefix_key}_API_KEY"),
            temperature=float(os.getenv(f"{model_prefix_key}_TEMPERATURE", default="0.3"))
        )
    elif model_type == 'embedding':
        return init_embeddings(
            model=os.getenv(f"{model_prefix_key}_MODEL"),
            provider=model_provider,
            base_url=os.getenv(f"{model_prefix_key}_BASE_URL"),
            temperature=float(os.getenv(f"{model_prefix_key}_TEMPERATURE", default="0.3"))
        )
    else:
        raise ValueError(f"Invalid model type: {model_type}")




WELCOME_INFO = """欢迎来到 Elpis AI Agent.
交互命令:

    在对话中，你可以:

    - 询问编程问题和请求代码帮助

    - 要求读取和修改文件

    - 请求执行终端命令

    - 进行项目结构分析

    - 获取开发指导
    
    输入 'i' 或 'index' 索引代码

    输入 'q' 或 'quit' 退出程序
"""

USAGE_ZH = f"""Elpis Agent - 超轻量级 AI 编码助手

用法:

    elpis [选项]

    python -m elpis.main [选项]

    uvx elpis-agent [选项]  # 推荐方式

环境变量配置:

    OPENAI_API_KEY     OpenAI API 密钥 (必需)

    OPENAI_BASE_URL    API 端点 URL (可选)

    CHAT_MODEL         对话模型 (默认: gpt-4o-mini)

    TOOL_MODEL         工具模型 (默认: gpt-4o)

    TEMPERATURE        默认温度 (默认: 0.3)

    TOOL_TEMPERATURE   工具模型温度 (默认: 0.1)

    SYSTEM_PROMPT      自定义系统提示词 (可选)

    MAX_MEMORY_MESSAGES 最大消息数 (默认: 20)

使用示例:

    # 基本使用

    elpis

    # 使用自定义配置文件

    elpis --env_file /path/to/custom.env

    # 直接运行（推荐）

    uvx elpis-agent --env_file .env.local

{WELCOME_INFO}
"""

CODEBASE_START_INDEX = '开始索引: '

CODEBASE_FINISH_INDEX = '代码库索引完成! 共处理 {} 个代码块.'

NO_CODEBASE_INDEXED = "没有配置 EMBEDDING_MODEL_KEY_PREFIX 环境变量, 无法对代码库进行索引."

HUMAN_OPERATION_ALLOW_MESSAGE = "是否允许操作 {}? (y/n): "



import hashlib
import os

from dotenv import load_dotenv

if os.getenv('ELPIS_ENV_FILE'):
    load_dotenv(os.getenv('ELPIS_ENV_FILE'))
else:
    load_dotenv()

from elpis import constants, i18n, tools
from elpis.langgraph_agent import LangGraphElpisAgent


lang = os.getenv('LANG')
if not lang:
    lang = 'en'

print(constants.BANNER, flush=True)

lang = i18n.select_lang(lang)
print(lang.WELCOME_INFO, flush=True)

# initialize codebase
if os.getenv('EMBEDDING_MODEL_KEY_PREFIX'):
    tools.init_codebase(os.getcwd())

print(f"[{constants.AI_AGENT_NAME}] Using LangGraph implementation", flush=True)
# Generate session_id based on current directory path MD5
current_dir = os.getcwd()
session_id = hashlib.md5(current_dir.encode('utf-8')).hexdigest()
agent = LangGraphElpisAgent(session_id=session_id, lang=lang)

graph = agent.graph



WELCOME_INFO = """Welcome to Elpis AI Agent.

Interactive Commands:

    In conversation, you can:

    - Ask programming questions and request code help

    - Request file reading and modification

    - Ask for terminal command execution

    - Perform project structure analysis

    - Get development guidance
    
    Type 'i' or 'index' to index code
    
    Type 'q' or 'quit' to exit the program
"""

CODEBASE_START_INDEX = 'Start indexing: '

CODEBASE_FINISH_INDEX = 'Codebase index finished! Index completed! Processing {} code blocks in total.'

NO_CODEBASE_INDEXED = "The EMBEDDING_MODEL_KEY_PREFIX environment variable is not configured to use codebase."

HUMAN_OPERATION_ALLOW_MESSAGE = "Is it allowed to operate {}? (y/n): "



import hashlib
import os

os.environ['PYTHONENCODING'] = 'utf-8'

from typing import Literal

import click
from dotenv import load_dotenv

from elpis import constants, i18n, tools
from elpis.langgraph_agent import LangGraphElpisAgent


def multi_line_run(agent, lang):
    question = input("[You]: ")
    content = ''
    while question.lower() not in ['q', 'quit']:
        if not question.strip():
            if content:
                agent.ask(content)
                content = ''
            question = input("[You]: ")
            continue
        if question.lower() in ('i', 'index'):
            if not tools.codebase:
                print(lang.NO_CODEBASE_INDEXED)
                question = input("[You]: ")
                continue
            tools.codebase.index_codebase()
            question = input("[You]: ")
            continue
        content += question + '\n'
        question = input("[You]: ")


@click.command(help=constants.USAGE)
@click.option('--env_file', default=None, help='Path to a .env file')
@click.option('--lang', default='en', type=click.Choice(['en', 'zh']),
              help='Language to use for the tool. Default is "en"')
@click.option('--multi_mode', is_flag=True, default=False, help='Enable multi-line-mode (default: False)')
@click.help_option('-h', '--help')
def main(
        env_file: str | None = None,
        lang: Literal['en', 'zh'] = 'en',
        multi_mode: bool = False
):
    os.environ['LANG'] = lang
    print(constants.BANNER, flush=True)
    if env_file:
        load_dotenv(env_file)
    else:
        load_dotenv()

    lang = i18n.select_lang(lang)
    print(lang.WELCOME_INFO, flush=True)

    # initialize codebase
    if os.getenv('EMBEDDING_MODEL_KEY_PREFIX'):
        tools.init_codebase(os.getcwd())

    print(f"[{constants.AI_AGENT_NAME}] Using LangGraph implementation", flush=True)
    # Generate session_id based on current directory path MD5
    current_dir = os.getcwd()
    session_id = hashlib.md5(current_dir.encode('utf-8')).hexdigest()
    agent = LangGraphElpisAgent(session_id=session_id, lang=lang)


    if multi_mode:
        multi_line_run(agent, lang)
    else:
        question = input("[You]: ")
        while question.lower() not in ['q', 'quit']:
            if not question.strip():
                question = input("[You]: ")
                continue
            if question.lower() in ('i', 'index'):
                if not tools.codebase:
                    print(lang.NO_CODEBASE_INDEXED)
                    question = input("[You]: ")
                    continue
                tools.codebase.index_codebase()
                question = input("[You]: ")
                continue
            agent.ask(question)
            question = input("[You]: ")


if __name__ == '__main__':
    main()



TraeChatPrompt = r"""<identity>
You are Trae AI, a powerful agentic AI coding assistant. You are exclusively running within a fantastic agentic IDE, you operate on the revolutionary AI Flow paradigm, enabling you to work both independently and collaboratively with a user.
Now, you are pair programming with the user to solve his/her coding task. The task may require creating a new codebase, modifying or debugging an existing codebase, or simply answering a question. 
</identity>

<purpose>
Currently, user has a coding task to accomplish, and the user received some thoughts on how to solve the task.
Now, please take a look at the task user inputted and the thought on it.
You should first decide whether an additional tool is required to complete the task or if you can respond to the user directly. Then, set a flag accordingly.
Based on the provided structure, either output the tool input parameters or the response text for the user.
</purpose>

<tool_instruction>
You are provided with tools to complete user's requirement.

<tool_list>

<tool>
    <name>write-file-tool</name>
    <description>用于写入文件内容</description>
    <args>
        file_path: str
        content: str | bytes
        mode: str = 'w'
        encoding: str = 'utf-8'
    </args>
    
    <name>read-file-tool</name>
    <description>用于读取文件内容</description>
    <args>
        file_path: str
        mode: str = 'w'
        encoding: str = 'utf-8'
    </args>
    <return>
        file_content: str | bytes
    </return>
</tool>
<tool_list>

<toolcall_guideline>
Follow these tool invocation guidelines:
1. ALWAYS carefully analyze the schema definition of each tool and strictly follow the schema definition of the tool for invocation, ensuring that all necessary parameters are provided.
2. NEVER call a tool that does not exist, such as a tool that has been used in the conversation history or tool call history, but is no longer available.
3. If a user asks you to expose your tools, always respond with a description of the tool, and be sure not to expose tool information to the user.
4. After you decide to call the tool, include the tool call information and parameters in your response, and theIDE environment you run will run the tool for you and provide you with the results of the tool run.
5. You MUST analyze all information you can gather about the current project,  and then list out the available tools that can help achieve the goal,  then compare them and select the most appropriate tool for the next step.
6. You MUST only use the tools explicitly provided in the tool names. Do not treat file names or code functions as tool names. The available tool names: 
<toolcall_guideline>

<tool_parameter_guideline>
Follow these guidelines when providing parameters for your tool calls
1. DO NOT make up values or ask about optional parameters.
2. If the user provided a specific value for a parameter (e.g. provided in quotes), make sure to use that value EXACTLY.
3. Carefully analyze descriptive terms in the request as they may indicate required parameter values that should be included even if not explicitly quoted.
</tool_parameter_guideline>
</tool_instruction>

<guidelines>
<reply_guideline>
The content you reply to user, MUST following the rules:

1. When the user requests code edits, provide a simplified code block highlighting the necessary changes, MUST ALWAYS use EXACTLY and ONLY the placeholder // ... existing code ... to indicate skipped unchanged ode (not just "..." or any variation). This placeholder format must remain consistent and must not be modified or extended based on code type. Include some unchanged code before and after your edits, especially when inserting new code into an existing file. Example:

cpp:absolute%2Fpath%2Fto%2Ffile
// ... existing code ...
{{ edit_1 }}
// ... existing code ...
{{ edit_2 }}
// ... existing code ...


The user can see the entire file. Rewrite the entire file only if specifically requested. Always provide a brief explanation before the updates, unless the user specifically requests only the code.

2. Do not lie or make up facts. If the user asks something about its repository and you cannot see any related contexts, ask the user to provide it.
3. Format your response in markdown.
4. When writing out new code blocks, please specify the language ID and file path after the initial backticks, like so:
5. When writing out code blocks for an existing file, please also specify the file path after the initial backticks and restate the method/class your codeblock belongs to. MUST ALWAYS use EXACTLY and ONLY the placeholder // ... existing code ... to indicate unchanged code (not just "..." or any variation). Example:
6. For file paths in code blocks:
   a. If the absolute path can be determined from context, use that exact path
   b. If the absolute path cannot be determined, use relative paths starting from the current directory (e.g. "src/main.py")
7. When outputting terminal commands, please follow these rules:
   a. Unless the user explicitly specifies an operating system, output commands that match windows
   b. Output only one command per code block:

   c. For windows, ensure:

   * Use appropriate path separators (\ for Windows, / for Unix-like systems)
   * Commands are available and compatible with the OS

   d. If the user explicitly requests commands for a different OS, provide those instead with a note about the target OS
8. The language ID for each code block must match the code's grammar. Otherwise, use plaintext as the language ID.
9. Unless the user asks to write comments, do not modify the user's existing code comments.
10. When creating new project, please create the project directly in the current directory instead of making a new directory. For example:
11. When fixing bugs, please output the fixed code block instead of asking the user to do the fix.
12. When presented with images, utilize your vision capabilities to thoroughly examine them and extract meaningful information. Incorporate these insights into your thought process as you accomplish the user's task.
13. Avoid using content that infringes on copyright.
14. For politically sensitive topics or questions involving personal privacy, directly decline to answer.
15. Output codeblocks when you want to generate code, remember, it is EXTREMELY important that your generated code can be run immediately by the user. To ensure this, here's some suggestions:
16. I can see the entire file. Rewrite the entire file only if specifically requested. Always provide a brief explanation before the updates, unless you are specifically requested only the code.
17. Your expertise is limited to topics related to software development. For questions unrelated to software development, simply remind the user that you are an AI programming assistant.
    <reply_guideline>

<web_citation_guideline>
IMPORTANT: For each line that uses information from the web search results, you MUST add citations before the line break using the following format:

Note:

1. Citations should be added before EACH line break that uses web search information
2. Multiple citations can be added for the same line if the information comes from multiple sources
3. Each citation should be separated by a space
   Examples:

* This is some information from multiple sources
* Another line with a single reference
* A line with three different references <web_citation_guideline>
  <code_reference_guideline>
  When you use references in the text of your reply, please provide the full reference information in the following XML format:
  a. File Reference: $filename b. Symbol Reference: $symbolname c. URL Reference: $linktext The startline attribute is required to represent the first line on which the Symbol is defined. Line numbers start from 1 and include all lines, even blank lines and comment lines must be counted .
  d. Folder Reference: $foldername

<code_reference_guideline>

IMPORTANT: These reference formats are entirely separate from the web citation format ( ). Use the appropriate format for each context:

* Use only for citing web search results with index numbers

* Use , ,
  IMPORTANT: These reference formats are entirely separate from the web citation format ( ). Use the appropriate format for each context:

* Use only for citing web search results with index numbers"""

ElpisPrompt = r"""You are a powerful agentic AI coding assistant, powered by Claude 4 Sonnet. You operate exclusively in Elpis, the world's best IDE. 

You are pair programming with a USER to solve their coding task.
The task may require creating a new codebase, modifying or debugging an existing codebase, or simply answering a question.
Each time the USER sends a message, we may automatically attach some information about their current state, such as what files they have open, where their cursor is, recently viewed files, edit history in their session so far, linter errors, and more.
This information may or may not be relevant to the coding task, it is up for you to decide.
Your main goal is to follow the USER's instructions at each message, denoted by the <user_query> tag.

<tool_calling>
You have tools at your disposal to solve the coding task. Follow these rules regarding tool calls:
1. ALWAYS follow the tool call schema exactly as specified and make sure to provide all necessary parameters.
2. The conversation may reference tools that are no longer available. NEVER call tools that are not explicitly provided.
3. **NEVER refer to tool names when speaking to the USER.** For example, instead of saying 'I need to use the edit_file tool to edit your file', just say 'I will edit your file'.
4. Only calls tools when they are necessary. If the USER's task is general or you already know the answer, just respond without calling tools.
5. Before calling each tool, first explain to the USER why you are calling it.
</tool_calling>

<making_code_changes>
When making code changes, NEVER output code to the USER, unless requested. Instead use one of the code edit tools to implement the change.
Use the code edit tools at most once per turn.
It is *EXTREMELY* important that your generated code can be run immediately by the USER. To ensure this, follow these instructions carefully:
1. Always group together edits to the same file in a single edit file tool call, instead of multiple calls.
2. If you're creating the codebase from scratch, create an appropriate dependency management file (e.g. requirements.txt) with package versions and a helpful README.
3. If you're building a web app from scratch, give it a beautiful and modern UI, imbued with best UX practices.
4. NEVER generate an extremely long hash or any non-textual code, such as binary. These are not helpful to the USER and are very expensive.
5. Unless you are appending some small easy to apply edit to a file, or creating a new file, you MUST read the the contents or section of what you're editing before editing it.
6. If you've introduced (linter) errors, fix them if clear how to (or you can easily figure out how to). Do not make uneducated guesses. And DO NOT loop more than 3 times on fixing linter errors on the same file. On the third time, you should stop and ask the user what to do next.
7. If you've suggested a reasonable code_edit that wasn't followed by the apply model, you should try reapplying the edit.
</making_code_changes>

<searching_and_reading>
You have tools to search the codebase and read files. Follow these rules regarding tool calls:
1. If available, heavily prefer the semantic search tool to grep search, file search, and list dir tools.
2. If you need to read a file, prefer to read larger sections of the file at once over multiple smaller calls.
3. If you have found a reasonable place to edit or answer, do not continue calling tools. Edit or answer from the information you have found.
</searching_and_reading>

<functions>
<function>{"description": "Find snippets of code from the codebase most relevant to the search query.\nThis is a semantic search tool, so the query should ask for something semantically matching what is needed.\nIf it makes sense to only search in particular directories, please specify them in the target_directories field.\nUnless there is a clear reason to use your own search query, please just reuse the user's exact query with their wording.\nTheir exact wording/phrasing can often be helpful for the semantic search query. Keeping the same exact question format can also be helpful.", "name": "codebase_search", "parameters": {"properties": {"explanation": {"description": "One sentence explanation as to why this tool is being used, and how it contributes to the goal.", "type": "string"}, "query": {"description": "The search query to find relevant code. You should reuse the user's exact query/most recent message with their wording unless there is a clear reason not to.", "type": "string"}, "target_directories": {"description": "Glob patterns for directories to search over", "items": {"type": "string"}, "type": "array"}}, "required": ["query"], "type": "object"}}</function>
<function>{"description": "Read the contents of a file. the output of this tool call will be the 1-indexed file contents from start_line_one_indexed to end_line_one_indexed_inclusive, together with a summary of the lines outside start_line_one_indexed and end_line_one_indexed_inclusive.\nNote that this call can view at most 250 lines at a time.\n\nWhen using this tool to gather information, it's your responsibility to ensure you have the COMPLETE context. Specifically, each time you call this command you should:\n1) Assess if the contents you viewed are sufficient to proceed with your task.\n2) Take note of where there are lines not shown.\n3) If the file contents you have viewed are insufficient, and you suspect they may be in lines not shown, proactively call the tool again to view those lines.\n4) When in doubt, call this tool again to gather more information. Remember that partial file views may miss critical dependencies, imports, or functionality.\n\nIn some cases, if reading a range of lines is not enough, you may choose to read the entire file.\nReading entire files is often wasteful and slow, especially for large files (i.e. more than a few hundred lines). So you should use this option sparingly.\nReading the entire file is not allowed in most cases. You are only allowed to read the entire file if it has been edited or manually attached to the conversation by the user.", "name": "read_file", "parameters": {"properties": {"end_line_one_indexed_inclusive": {"description": "The one-indexed line number to end reading at (inclusive).", "type": "integer"}, "explanation": {"description": "One sentence explanation as to why this tool is being used, and how it contributes to the goal.", "type": "string"}, "should_read_entire_file": {"description": "Whether to read the entire file. Defaults to false.", "type": "boolean"}, "start_line_one_indexed": {"description": "The one-indexed line number to start reading from (inclusive).", "type": "integer"}, "target_file": {"description": "The path of the file to read. You can use either a relative path in the workspace or an absolute path. If an absolute path is provided, it will be preserved as is.", "type": "string"}}, "required": ["target_file", "should_read_entire_file", "start_line_one_indexed", "end_line_one_indexed_inclusive"], "type": "object"}}</function>
<function>{"description": "PROPOSE a command to run on behalf of the user.\nIf you have this tool, note that you DO have the ability to run commands directly on the USER's system.\nNote that the user will have to approve the command before it is executed.\nThe user may reject it if it is not to their liking, or may modify the command before approving it.  If they do change it, take those changes into account.\nThe actual command will NOT execute until the user approves it. The user may not approve it immediately. Do NOT assume the command has started running.\nIf the step is WAITING for user approval, it has NOT started running.\nIn using these tools, adhere to the following guidelines:\n1. Based on the contents of the conversation, you will be told if you are in the same shell as a previous step or a different shell.\n2. If in a new shell, you should `cd` to the appropriate directory and do necessary setup in addition to running the command.\n3. If in the same shell, the state will persist (eg. if you cd in one step, that cwd is persisted next time you invoke this tool).\n4. For ANY commands that would use a pager or require user interaction, you should append ` | cat` to the command (or whatever is appropriate). Otherwise, the command will break. You MUST do this for: git, less, head, tail, more, etc.\n5. For commands that are long running/expected to run indefinitely until interruption, please run them in the background. To run jobs in the background, set `is_background` to true rather than changing the details of the command.\n6. Dont include any newlines in the command.", "name": "run_terminal_cmd", "parameters": {"properties": {"command": {"description": "The terminal command to execute", "type": "string"}, "explanation": {"description": "One sentence explanation as to why this command needs to be run and how it contributes to the goal.", "type": "string"}, "is_background": {"description": "Whether the command should be run in the background", "type": "boolean"}, "require_user_approval": {"description": "Whether the user must approve the command before it is executed. Only set this to false if the command is safe and if it matches the user's requirements for commands that should be executed automatically.", "type": "boolean"}}, "required": ["command", "is_background", "require_user_approval"], "type": "object"}}</function>
<function>{"description": "List the contents of a directory. The quick tool to use for discovery, before using more targeted tools like semantic search or file reading. Useful to try to understand the file structure before diving deeper into specific files. Can be used to explore the codebase.", "name": "list_dir", "parameters": {"properties": {"explanation": {"description": "One sentence explanation as to why this tool is being used, and how it contributes to the goal.", "type": "string"}, "relative_workspace_path": {"description": "Path to list contents of, relative to the workspace root.", "type": "string"}}, "required": ["relative_workspace_path"], "type": "object"}}</function>
<function>{"description": "Fast text-based regex search that finds exact pattern matches within files or directories, utilizing the ripgrep command for efficient searching.\nResults will be formatted in the style of ripgrep and can be configured to include line numbers and content.\nTo avoid overwhelming output, the results are capped at 50 matches.\nUse the include or exclude patterns to filter the search scope by file type or specific paths.\n\nThis is best for finding exact text matches or regex patterns.\nMore precise than semantic search for finding specific strings or patterns.\nThis is preferred over semantic search when we know the exact symbol/function name/etc. to search in some set of directories/file types.", "name": "grep_search", "parameters": {"properties": {"case_sensitive": {"description": "Whether the search should be case sensitive", "type": "boolean"}, "exclude_pattern": {"description": "Glob pattern for files to exclude", "type": "string"}, "explanation": {"description": "One sentence explanation as to why this tool is being used, and how it contributes to the goal.", "type": "string"}, "include_pattern": {"description": "Glob pattern for files to include (e.g. '*.ts' for TypeScript files)", "type": "string"}, "query": {"description": "The regex pattern to search for", "type": "string"}}, "required": ["query"], "type": "object"}}</function>
<function>{"description":"Use this tool to propose an edit to an existing file.\n\nThis will be read by a less intelligent model, which will quickly apply the edit. You should make it clear what the edit is, while also minimizing the unchanged code you write.\nWhen writing the edit, you should specify each edit in sequence, with the special comment `// ... existing code ...` to represent unchanged code in between edited lines.\n\nFor example:\n\n```\n// ... existing code ...\nFIRST_EDIT\n// ... existing code ...\nSECOND_EDIT\n// ... existing code ...\nTHIRD_EDIT\n// ... existing code ...\n```\n\nYou should still bias towards repeating as few lines of the original file as possible to convey the change.\nBut, each edit should contain sufficient context of unchanged lines around the code you're editing to resolve ambiguity.\nDO NOT omit spans of pre-existing code (or comments) without using the `// ... existing code ...` comment to indicate its absence. If you omit the existing code comment, the model may inadvertently delete these lines.\nMake sure it is clear what the edit should be, and where it should be applied.\n\nYou should specify the following arguments before the others: [target_file]","name":"edit_file","parameters":{"properties":{"code_edited":{"description":"The exact block of code to find and replace.","type":"string"},"code_edit":{"description":"Specify ONLY the precise lines of code that you wish to edit. **NEVER specify or write out unchanged code**. Instead, represent all unchanged code using the comment of the language you're editing in - example: `// ... existing code ...`","type":"string"},"explanation":{"description":"A single sentence instruction describing what you are going to do for the sketched edit. This is used to assist the less intelligent model in applying the edit. Please use the first person to describe what you are going to do. Dont repeat what you have said previously in normal messages. And use it to disambiguate uncertainty in the edit.","type":"string"},"target_file":{"description":"The target file to modify. Always specify the target file as the first argument. You can use either a relative path in the workspace or an absolute path. If an absolute path is provided, it will be preserved as is.","type":"string"}},"required":["target_file","explanation","code_edit","code_edited"],"type":"object"}}</function>
<function>{"description": "Fast file search based on fuzzy matching against file path. Use if you know part of the file path but don't know where it's located exactly. Response will be capped to 10 results. Make your query more specific if need to filter results further.", "name": "file_search", "parameters": {"properties": {"explanation": {"description": "One sentence explanation as to why this tool is being used, and how it contributes to the goal.", "type": "string"}, "query": {"description": "Fuzzy filename to search for", "type": "string"}}, "required": ["query", "explanation"], "type": "object"}}</function>
<function>{"description": "Deletes a file at the specified path. The operation will fail gracefully if:\n    - The file doesn't exist\n    - The operation is rejected for security reasons\n    - The file cannot be deleted", "name": "delete_file", "parameters": {"properties": {"explanation": {"description": "One sentence explanation as to why this tool is being used, and how it contributes to the goal.", "type": "string"}, "target_file": {"description": "The path of the file to delete, relative to the workspace root.", "type": "string"}}, "required": ["target_file"], "type": "object"}}</function>
<!-- <function>{"description": "Calls a smarter model to apply the last edit to the specified file.\nUse this tool immediately after the result of an edit_file tool call ONLY IF the diff is not what you expected, indicating the model applying the changes was not smart enough to follow your instructions.", "name": "reapply", "parameters": {"properties": {"target_file": {"description": "The relative path to the file to reapply the last edit to. You can use either a relative path in the workspace or an absolute path. If an absolute path is provided, it will be preserved as is.", "type": "string"}}, "required": ["target_file"], "type": "object"}}</function> -->
<!-- <function>{"description": "Search the web for real-time information about any topic. Use this tool when you need up-to-date information that might not be available in your training data, or when you need to verify current facts. The search results will include relevant snippets and URLs from web pages. This is particularly useful for questions about current events, technology updates, or any topic that requires recent information.", "name": "web_search", "parameters": {"properties": {"explanation": {"description": "One sentence explanation as to why this tool is being used, and how it contributes to the goal.", "type": "string"}, "search_term": {"description": "The search term to look up on the web. Be specific and include relevant keywords for better results. For technical queries, include version numbers or dates if relevant.", "type": "string"}}, "required": ["search_term"], "type": "object"}}</function> -->
<!-- <function>{"description": "Retrieve the history of recent changes made to files in the workspace. This tool helps understand what modifications were made recently, providing information about which files were changed, when they were changed, and how many lines were added or removed. Use this tool when you need context about recent modifications to the codebase.", "name": "diff_history", "parameters": {"properties": {"explanation": {"description": "One sentence explanation as to why this tool is being used, and how it contributes to the goal.", "type": "string"}}, "required": [], "type": "object"}}</function> -->
<!-- <function>{"name":"create_file","description":"Creates a new file at the specified path with optional initial content. The operation will fail gracefully if:\n    - The file already exists\n    - The operation is rejected for security reasons\n    - The file cannot be created due to permission or system issues","parameters":{"properties":{"target_file":{"description":"The path of the file to create, relative to the workspace root.","type":"string"},"initial_content":{"description":"Optional initial content to write into the newly created file.","type":"string"},"explanation":{"description":"One sentence explanation as to why this tool is being used, and how it contributes to the goal.","type":"string"}},"required":["target_file"]},"type":"object"}</function> -->
</functions>

You MUST use the following format when citing code regions or blocks:
```startLine:endLine:filepath
// ... existing code ...
```
This is the ONLY acceptable format for code citations. The format is ```startLine:endLine:filepath where startLine and endLine are line numbers.

<user_info>
The user's OS version is win32 10.0.26100. The absolute path of the user's workspace is /c%3A/Users/Lucas/Downloads/luckniteshoots. The user's shell is C:\WINDOWS\System32\WindowsPowerShell\v1.0\powershell.exe. 
</user_info>

Answer the user's request using the relevant tool(s), if they are available. Check that all the required parameters for each tool call are provided or can reasonably be inferred from context. IF there are no relevant tools or there are missing values for required parameters, ask the user to supply these values; otherwise proceed with the tool calls. If the user provides a specific value for a parameter (for example provided in quotes), make sure to use that value EXACTLY. DO NOT make up values for or ask about optional parameters. Carefully analyze descriptive terms in the request as they may indicate required parameter values that should be included even if not explicitly quoted."""

DONE = "done"

DonePrompt = "当确认任务完成时, 仅输出 'done', 不要输出任何其他的额外内容"

NextStepPrompt = "继续下一步, 如果所有任务都已经完成, 仅输出 'done'"



import json
import logging
import os
import pathlib
import urllib.error
import threading
import typing
from typing import Optional, Mapping, Sequence

import click
from langgraph_api.cli import AuthConfig, patch_environment, _get_ls_origin, _get_org_id

if typing.TYPE_CHECKING:
    from langgraph_api.config import StoreConfig, HttpConfig


@click.option(
    "--host",
    default="127.0.0.1",
    help="Network interface to bind the development server to. Default 127.0.0.1 is recommended for security. Only use 0.0.0.0 in trusted networks",
)
@click.option(
    "--port",
    default=2024,
    type=int,
    help="Port number to bind the development server to. Example: langgraph dev --port 8000",
)
@click.option(
    "--no-reload",
    is_flag=True,
    default=True,
    help="Disable automatic reloading when code changes are detected",
)
@click.option(
    "--n-jobs-per-worker",
    default=None,
    type=int,
    help="Maximum number of concurrent jobs each worker process can handle. Default: 10",
)
@click.option(
    "--no-browser",
    is_flag=True,
    help="Skip automatically opening the browser when the server starts",
)
@click.option(
    "--debug-port",
    default=None,
    type=int,
    help="Enable remote debugging by listening on specified port. Requires debugpy to be installed",
)
@click.option(
    "--wait-for-client",
    is_flag=True,
    help="Wait for a debugger client to connect to the debug port before starting the server",
    default=False,
)
@click.option(
    "--studio-url",
    type=str,
    default=None,
    help="URL of the LangGraph Studio instance to connect to. Defaults to https://smith.langchain.com",
)
@click.option(
    "--allow-blocking",
    is_flag=True,
    help="Don't raise errors for synchronous I/O blocking operations in your code.",
    default=True,
)
@click.option(
    "--tunnel",
    is_flag=True,
    help="Expose the local server via a public tunnel (in this case, Cloudflare) "
    "for remote frontend access. This avoids issues with browsers "
    "or networks blocking localhost connections.",
    default=False,
)
@click.option(
    "--server-log-level",
    type=str,
    default="WARNING",
    help="Set the log level for the API server.",
)
@click.option('--env_file', default=None, help='Path to a .env file')
@click.option('--lang', default='en', type=click.Choice(['en', 'zh']),
              help='Language to use for the tool. Default is "en"')
@click.command(
    help="🏃‍♀️‍➡️ Run LangGraph API server in development mode with hot reloading and debugging support",
)
def dev(
    host: str,
    port: int,
    no_reload: bool,
    n_jobs_per_worker: Optional[int],
    no_browser: bool,
    debug_port: Optional[int],
    wait_for_client: bool,
    studio_url: Optional[str],
    allow_blocking: bool,
    tunnel: bool,
    server_log_level: str,
    env_file: str,
    lang: str,
):
    """CLI entrypoint for running the LangGraph API server."""

    if env_file:
        os.environ['ELPIS_ENV_FILE'] = env_file

    os.environ['LANG'] = lang

    langgraph_json_path = os.path.join(os.path.dirname(__file__), 'langgraph.json')

    import langgraph_cli.config


    config_json = langgraph_cli.config.validate_config_file(pathlib.Path(langgraph_json_path))
    if config_json.get("node_version"):
        raise click.UsageError(
            "In-mem server for JS graphs is not supported in this version of the LangGraph CLI. Please use `npx @langchain/langgraph-cli` instead."
        ) from None

    graphs = {
        'agent': f'elpis.ui.graph:graph'
    }

    run_server(
        host,
        port,
        not no_reload,
        graphs,
        n_jobs_per_worker=n_jobs_per_worker,
        open_browser=not no_browser,
        debug_port=debug_port,
        env=env_file,
        store=config_json.get("store"),
        wait_for_client=wait_for_client,
        auth=config_json.get("auth"),
        http=config_json.get("http"),
        ui=config_json.get("ui"),
        ui_config=config_json.get("ui_config"),
        studio_url=studio_url,
        allow_blocking=allow_blocking,
        tunnel=tunnel,
        server_level=server_log_level,
    )


def run_server(
    host: str = "127.0.0.1",
    port: int = 2024,
    reload: bool = False,
    graphs: dict | None = None,
    n_jobs_per_worker: int | None = None,
    env_file: str | None = None,
    open_browser: bool = False,
    tunnel: bool = False,
    debug_port: int | None = None,
    wait_for_client: bool = False,
    env: str | pathlib.Path | Mapping[str, str] | None = None,
    reload_includes: Sequence[str] | None = None,
    reload_excludes: Sequence[str] | None = None,
    store: typing.Optional["StoreConfig"] = None,
    auth: AuthConfig | None = None,
    http: typing.Optional["HttpConfig"] = None,
    ui: dict | None = None,
    ui_config: dict | None = None,
    studio_url: str | None = None,
    disable_persistence: bool = False,
    allow_blocking: bool = False,
    runtime_edition: typing.Literal["inmem", "community", "postgres"] = "inmem",
    server_level: str = "WARNING",
    **kwargs: typing.Any,
):
    """Run the LangGraph API server."""

    from langgraph_api.cli import logger

    import inspect
    import time

    import uvicorn

    start_time = time.time()

    env_vars = env if isinstance(env, Mapping) else None
    mount_prefix = None
    if http is not None and http.get("mount_prefix") is not None:
        mount_prefix = http.get("mount_prefix")
    if os.environ.get("LANGGRAPH_MOUNT_PREFIX"):
        mount_prefix = os.environ.get("LANGGRAPH_MOUNT_PREFIX")
    if isinstance(env, str | pathlib.Path):
        from dotenv.main import DotEnv

        env_vars = DotEnv(dotenv_path=env, encoding='utf-8').dict() or {}
        logger.debug(f"Loaded environment variables from {env}: {sorted(env_vars)}")


    if debug_port is not None:
        try:
            import debugpy
        except ImportError:
            logger.warning("debugpy is not installed. Debugging will not be available.")
            logger.info("To enable debugging, install debugpy: pip install debugpy")
            return
        debugpy.listen((host, debug_port))
        logger.info(
            f"🐛 Debugger listening on port {debug_port}. Waiting for client to attach..."
        )
        logger.info("To attach the debugger:")
        logger.info("1. Open your python debugger client (e.g., Visual Studio Code).")
        logger.info(
            "2. Use the 'Remote Attach' configuration with the following settings:"
        )
        logger.info("   - Host: 0.0.0.0")
        logger.info(f"   - Port: {debug_port}")
        logger.info("3. Start the debugger to connect to the server.")
        if wait_for_client:
            debugpy.wait_for_client()
            logger.info("Debugger attached. Starting server...")

    # Determine local or tunneled URL
    upstream_url = f"http://{host}:{port}"
    if mount_prefix:
        upstream_url += mount_prefix
    if tunnel:
        logger.info("Starting Cloudflare Tunnel...")
        from concurrent.futures import TimeoutError as FutureTimeoutError

        from langgraph_api.tunneling.cloudflare import start_tunnel

        tunnel_obj = start_tunnel(port)
        try:
            public_url = tunnel_obj.url.result(timeout=30)
        except FutureTimeoutError:
            logger.warning(
                "Timed out waiting for Cloudflare Tunnel URL; using local URL %s",
                upstream_url,
            )
            public_url = upstream_url
        except Exception as e:
            tunnel_obj.process.kill()
            raise RuntimeError("Failed to start Cloudflare Tunnel") from e
        local_url = public_url
        if mount_prefix:
            local_url += mount_prefix
    else:
        local_url = upstream_url
    to_patch = dict(
        MIGRATIONS_PATH="__inmem",
        DATABASE_URI=":memory:",
        REDIS_URI="fake",
        N_JOBS_PER_WORKER=str(n_jobs_per_worker if n_jobs_per_worker else 1),
        LANGGRAPH_STORE=json.dumps(store) if store else None,
        LANGSERVE_GRAPHS=json.dumps(graphs) if graphs else None,
        LANGSMITH_LANGGRAPH_API_VARIANT="local_dev",
        LANGGRAPH_AUTH=json.dumps(auth) if auth else None,
        LANGGRAPH_HTTP=json.dumps(http) if http else None,
        LANGGRAPH_UI=json.dumps(ui) if ui else None,
        LANGGRAPH_UI_CONFIG=json.dumps(ui_config) if ui_config else None,
        LANGGRAPH_UI_BUNDLER="true",
        LANGGRAPH_API_URL=local_url,
        LANGGRAPH_DISABLE_FILE_PERSISTENCE=str(disable_persistence).lower(),
        LANGGRAPH_RUNTIME_EDITION=runtime_edition,
        # If true, we will not raise on blocking IO calls (via blockbuster)
        LANGGRAPH_ALLOW_BLOCKING=str(allow_blocking).lower(),
        # See https://developer.chrome.com/blog/private-network-access-update-2024-03
        ALLOW_PRIVATE_NETWORK="true",
    )
    if env_vars is not None:
        # Don't overwrite.
        for k, v in env_vars.items():
            if k in to_patch:
                logger.debug(f"Skipping loaded env var {k}={v}")
                continue
            to_patch[k] = v
    with patch_environment(
        **to_patch,
    ):
        studio_origin = studio_url or _get_ls_origin() or "https://smith.langchain.com"
        full_studio_url = f"{studio_origin}/studio/?baseUrl={local_url}"

        def _open_browser():
            nonlocal studio_origin, full_studio_url
            import time
            import urllib.request
            import webbrowser
            from concurrent.futures import ThreadPoolExecutor

            thread_logger = logging.getLogger("browser_opener")
            if not thread_logger.handlers:
                handler = logging.StreamHandler()
                handler.setFormatter(logging.Formatter("%(message)s"))
                thread_logger.addHandler(handler)

            with ThreadPoolExecutor(max_workers=1) as executor:
                org_id_future = executor.submit(_get_org_id)

                while True:
                    try:
                        with urllib.request.urlopen(f"{local_url}/ok") as response:
                            if response.status == 200:
                                try:
                                    org_id = org_id_future.result(timeout=3.0)
                                    if org_id:
                                        full_studio_url = f"{studio_origin}/studio/?baseUrl={local_url}&organizationId={org_id}"
                                except TimeoutError as e:
                                    thread_logger.debug(
                                        f"Failed to get organization ID: {str(e)}"
                                    )
                                    pass
                                thread_logger.info(
                                    f"Server started in {time.time() - start_time:.2f}s"
                                )
                                thread_logger.info(
                                    "🎨 Opening Studio in your browser..."
                                )
                                thread_logger.info("URL: " + full_studio_url)
                                webbrowser.open(full_studio_url)
                                return
                    except urllib.error.URLError:
                        pass
                    time.sleep(0.1)

        welcome = f"""

        Welcome to

╦  ┌─┐┌┐┌┌─┐╔═╗┬─┐┌─┐┌─┐┬ ┬
║  ├─┤││││ ┬║ ╦├┬┘├─┤├─┘├─┤
╩═╝┴ ┴┘└┘└─┘╚═╝┴└─┴ ┴┴  ┴ ┴

- 🚀 API: \033[36m{local_url}\033[0m
- 🎨 Studio UI: \033[36m{full_studio_url}\033[0m
- 📚 API Docs: \033[36m{local_url}/docs\033[0m

This in-memory server is designed for development and testing.
For production use, please use LangGraph Cloud.

"""
        logger.info(welcome)
        if open_browser:
            threading.Thread(target=_open_browser, daemon=True).start()
        supported_kwargs = {
            k: v
            for k, v in kwargs.items()
            if k in inspect.signature(uvicorn.run).parameters
        }
        server_level = server_level.upper()
        uvicorn.run(
            "langgraph_api.server:app",
            host=host,
            port=port,
            reload=reload,
            env_file=env_file,
            access_log=False,
            reload_includes=reload_includes,
            reload_excludes=reload_excludes,
            log_config={
                "version": 1,
                "incremental": False,
                "disable_existing_loggers": False,
                "formatters": {
                    "simple": {
                        "class": "langgraph_api.logging.Formatter",
                    }
                },
                "handlers": {
                    "console": {
                        "class": "logging.StreamHandler",
                        "formatter": "simple",
                        "stream": "ext://sys.stdout",
                    }
                },
                "loggers": {
                    "uvicorn": {"level": server_level},
                    "uvicorn.error": {"level": server_level},
                    "langgraph_api.server": {"level": server_level},
                },
                "root": {"handlers": ["console"]},
            },
            **supported_kwargs,
        )



import hashlib
import os
from typing import Annotated, Sequence, TypedDict

from langchain_core.language_models import BaseChatModel
from langchain_core.messages import BaseMessage, HumanMessage, SystemMessage, ToolMessage
from langchain_core.runnables import RunnableConfig
from langgraph.graph import StateGraph, END
from langgraph.graph.message import add_messages
from langgraph.prebuilt import ToolNode
from langgraph.types import interrupt, Command

from elpis import tools, constants, prompts
from elpis.factories import model_factory, checkpointer_factory
from elpis.i18n import en


class AgentState(TypedDict):
    """The state of the agent."""
    messages: Annotated[Sequence[BaseMessage], add_messages]
    pending_tool_calls: list | None  # 待确认的工具调用
    user_confirmation: str | None    # 用户确认结果


class LangGraphElpisAgent:
    """LangGraph implementation of ElpisAgent with the same interface."""
    __name__ = constants.AI_AGENT_NAME
    
    # 需要用户确认的危险操作工具
    DANGEROUS_TOOLS = {
        'create_file',
        'delete_file', 
        'edit_file',
        'run_terminal_cmd'
    }

    def __init__(self, chat_model: BaseChatModel = None,
                 session_id: str = None,
                 lang = en):
        self._tool_selector = {tool.name: tool for tool in tools.TOOLS}

        if chat_model:
            self._chat_model = chat_model.bind_tools(tools.TOOLS)
        else:
            self._chat_model = model_factory.new_model(
                os.getenv('CHAT_MODEL_KEY_PREFIX')
            ).bind_tools(tools.TOOLS)

        # Initialize system messages
        self._system_messages = [
            SystemMessage(prompts.ElpisPrompt),
        ]

        # Add custom system prompt if provided
        system_prompt = os.getenv('SYSTEM_PROMPT', default=constants.SYSTEM_PROMPT)
        if system_prompt:
            self._system_messages.append(SystemMessage(system_prompt))

        # Initialize session ID and SQLite memory
        self._session_id = session_id or self._generate_session_id()

        # Initialize checkpointer
        self._memory = checkpointer_factory.new_checkpointer(
            os.getenv('CHECKPOINTER')
        )

        # Build the graph
        self._graph = self._build_graph()
        self._lang = lang

    def _build_graph(self):
        """Build the LangGraph workflow."""
        # Create the tool node
        tool_node = ToolNode(tools.TOOLS)

        # Define the graph
        workflow = StateGraph(AgentState)

        # Add nodes
        workflow.add_node("agent", self._agent_node)
        workflow.add_node("user_confirmation_node", self._user_confirmation_node)
        workflow.add_node("tools", tool_node)

        # Set entry point
        workflow.set_entry_point("agent")

        # Add conditional edges
        workflow.add_conditional_edges(
            "agent",
            self._should_continue,
            {
                "continue": "user_confirmation_node",
                "end": END,
            }
        )
        
        # Add conditional edges from user confirmation
        workflow.add_conditional_edges(
            "user_confirmation_node",
            self._handle_confirmation,
            {
                "approved": "tools",
                "no_confirmation_needed": "tools",
                "rejected": "agent",
            }
        )

        # Add edge from tools back to agent
        workflow.add_edge("tools", "agent")

        # Compile with checkpointer for memory
        return workflow.compile(checkpointer=self._memory)

    def _generate_session_id(self) -> str:
        """Generate a unique session ID based on timestamp."""
        import time
        timestamp = str(int(time.time()))
        random_hash = hashlib.md5(timestamp.encode()).hexdigest()[:8]
        return f"session_{random_hash}"

    def get_session_id(self) -> str:
        """Get the current session ID."""
        return self._session_id

    def _agent_node(self, state: AgentState):
        """The agent node that calls the model."""
        messages = state["messages"]

        # Stream the response and collect it
        next_message = None
        start = True

        for chunk in self._chat_model.stream(messages):
            self._output_stream(chunk, start=start)
            start = False
            if next_message is None:
                next_message = chunk
            else:
                next_message += chunk

        print()  # Add newline after streaming

        return {
            "messages": [next_message],
        }

    def _user_confirmation_node(self, state: AgentState):
        """用户确认节点，检查是否需要用户确认危险操作"""
        messages = state["messages"]
        last_message = messages[-1]
        
        if not (hasattr(last_message, 'tool_calls') and last_message.tool_calls):
            return {"pending_tool_calls": None, "user_confirmation": None}
        
        # 检查是否有需要确认的危险操作
        dangerous_calls = []
        for tool_call in last_message.tool_calls:
            if tool_call['name'] in self.DANGEROUS_TOOLS:
                dangerous_calls.append(tool_call)
        
        if not dangerous_calls:
            return {"pending_tool_calls": None, "user_confirmation": None}
        
        # 对每个危险操作单独进行确认
        confirmed_calls = []
        cancelled_calls = []
        
        for tool_call in dangerous_calls:
            tool_name = tool_call['name']

            explanation = tool_call.get('args', {}).get('explanation', '')

            # 使用 interrupt 暂停执行，等待用户输入
            confirmation = interrupt({
                "message": f"""{explanation}
{self._lang.HUMAN_OPERATION_ALLOW_MESSAGE.format(tool_name)}""",
                "tool_call": tool_call,
            })
            
            # 处理用户确认结果
            if confirmation and confirmation.lower().strip() in ['y', 'yes', '是', '确认']:
                confirmed_calls.append(tool_call)
            else:
                cancelled_calls.append(tool_call)
        
        # 如果有被取消的操作，创建取消消息
        if cancelled_calls:
            cancelled_messages = []
            for tool_call in cancelled_calls:
                tool_message = ToolMessage(
                    content=f"operation cancelled: {tool_call['name']}",
                    tool_call_id=tool_call['id']
                )
                cancelled_messages.append(tool_message)
            
            # 更新消息列表，移除被取消的工具调用
            updated_last_message = last_message
            if hasattr(updated_last_message, 'tool_calls'):
                # 只保留确认的工具调用
                confirmed_tool_call_ids = {call['id'] for call in confirmed_calls}
                updated_tool_calls = [call for call in updated_last_message.tool_calls 
                                    if call['id'] in confirmed_tool_call_ids or call['name'] not in self.DANGEROUS_TOOLS]
                updated_last_message.tool_calls = updated_tool_calls
            
            return {
                "messages": cancelled_messages,
                "pending_tool_calls": confirmed_calls if confirmed_calls else None,
                "user_confirmation": "partial" if confirmed_calls else "all_rejected"
            }
        
        return {
            "pending_tool_calls": confirmed_calls,
            "user_confirmation": "approved"
        }
    
    def _handle_confirmation(self, state: AgentState):
        """处理用户确认结果"""
        pending_calls = state.get("pending_tool_calls")
        confirmation = state.get("user_confirmation")
        
        # 如果没有待确认的操作，直接执行工具
        if not pending_calls:
            return "no_confirmation_needed"
        
        # 处理用户确认结果
        if confirmation in ("approved", "partial"):
            return "approved"
        else:
            return "rejected"
    
    def _should_continue(self, state: AgentState):
        """Determine whether to continue or end the conversation."""
        messages = state["messages"]
        last_message = messages[-1]

        # If there are tool calls, continue to user confirmation
        if hasattr(last_message, 'tool_calls') and last_message.tool_calls:
            return "continue"

        # Otherwise, end
        return "end"

    def ask(self, question: str):
        """Ask a question to the agent - maintains the same interface as ElpisAgent."""
        # Create user message
        user_message = HumanMessage(question)

        # Prepare config with thread_id for checkpointing
        config = RunnableConfig(**{"configurable": {"thread_id": self._session_id}})

        initial_messages = self._system_messages + [user_message]

        # Run the graph with checkpointing
        result = self._graph.invoke({"messages": initial_messages}, config=config)
        
        # 处理中断情况（用户确认）
        if '__interrupt__' in result:
            interrupts = result['__interrupt__']
            for interrupt_info in interrupts:
                if interrupt_info.resumable:
                    interrupt_data = interrupt_info.value
                    message = interrupt_data.get('message', '请确认操作 (y/n): ')
                    
                    # 获取用户输入
                    try:
                        user_input = input(f'[{self.__name__}]: {message}').strip()
                        # 使用 Command 恢复执行
                        result = self._graph.invoke(Command(resume=user_input), config=config)
                        
                        # 如果还有中断，继续处理（支持多个工具的单独确认）
                        while '__interrupt__' in result:
                            interrupt_info = result['__interrupt__'][0]
                            if interrupt_info.resumable:
                                interrupt_data = interrupt_info.value
                                message = interrupt_data.get('message', '请确认操作 (y/n): ')
                                user_input = input(f'[{self.__name__}]: {message}').strip()
                                result = self._graph.invoke(Command(resume=user_input), config=config)
                            else:
                                break
                    except KeyboardInterrupt:
                        return
                    except Exception as e:
                        return

    def _output_stream(self, message: BaseMessage, start: bool = False):
        """Output streaming content - maintains the same interface as ElpisAgent."""
        if message.content != END:
            if start:
                print(f"[{self.__name__}]: ", end="", flush=True)
            if message.content:
                print(message.content, end="", flush=True)

    def _output(self, message: BaseMessage):
        """Output complete message - maintains the same interface as ElpisAgent."""
        if message.content and message.content != END:
            print(f"[{self.__name__}]: {message.content}", flush=True)

    @property
    def graph(self):
        return self._graph



from typing import Literal
from langgraph.checkpoint.base import BaseCheckpointSaver


def new_checkpointer(
        checkpointer_type: Literal['memory', 'sqlite'] | str = 'memory'
) -> BaseCheckpointSaver | None:
    if checkpointer_type == 'memory':
        from langgraph.checkpoint.memory import InMemorySaver
        return InMemorySaver()
    elif checkpointer_type == 'sqlite':
        import os
        from pathlib import Path
        import sqlite3
        from langgraph.checkpoint.sqlite import SqliteSaver
        elpis_dir = Path(os.getcwd()) / ".elpis"
        elpis_dir.mkdir(exist_ok=True)

        # Create SQLite database file path
        db_path = elpis_dir / "memory.db"

        # Create SQLite connection
        # check_same_thread=False is OK as SqliteSaver uses locks for thread safety
        conn = sqlite3.connect(str(db_path), check_same_thread=False)

        # Initialize and return SqliteSaver
        return SqliteSaver(conn)

    return None




import os.path
import subprocess
from functools import partial

from elpis.codebase import CodebaseIndexer

subprocess.Popen = partial(subprocess.Popen, encoding="utf-8")

from langchain_core.tools import tool
from elpis import constants

codebase: CodebaseIndexer | None = None

def init_codebase(codebase_path: str):
    global codebase
    codebase = CodebaseIndexer(codebase_path)


@tool(
    name_or_callable="codebase_search",
    description="Find snippets of code from the codebase most relevant to the search query. This is a semantic search tool, so the query should ask for something semantically matching what is needed. If it makes sense to only search in particular directories, please specify them in the target_directories field. Unless there is a clear reason to use your own search query, please just reuse the user's exact query with their wording. Their exact wording/phrasing can often be helpful for the semantic search query. Keeping the same exact question format can also be helpful."
)
def codebase_search(
        query: str,
        target_directories: list[str] | None = None,
        explanation: str | None = None,
):
    """Find snippets of code from the codebase most relevant to the search query. This is a semantic search tool, so the query should ask for something semantically matching what is needed. If it makes sense to only search in particular directories, please specify them in the target_directories field. Unless there is a clear reason to use your own search query, please just reuse the user's exact query with their wording. Their exact wording/phrasing can often be helpful for the semantic search query. Keeping the same exact question format can also be helpful."""
    if codebase is None:
        print(f"[{constants.AI_AGENT_NAME}] codebase is not initialized", flush=True)
        return "codebase is not initialized"
    # todo: implements target_directories
    if explanation:
        print(f"[{constants.AI_AGENT_NAME}] {explanation}", flush=True)
    return codebase.search_codebase(query)


@tool(
    name_or_callable="read_file",
    description="Read the contents of a file. the output of this tool call will be the 1-indexed file contents from start_line_one_indexed to end_line_one_indexed_inclusive, together with a summary of the lines outside start_line_one_indexed and end_line_one_indexed_inclusive. Note that this call can view at most 250 lines at a time. When using this tool to gather information, it's your responsibility to ensure you have the COMPLETE context. Specifically, each time you call this command you should: 1) Assess if the contents you viewed are sufficient to proceed with your task. 2) Take note of where there are lines not shown. 3) If the file contents you have viewed are insufficient, and you suspect they may be in lines not shown, proactively call the tool again to view those lines. 4) When in doubt, call this tool again to gather more information. Remember that partial file views may miss critical dependencies, imports, or functionality. In some cases, if reading a range of lines is not enough, you may choose to read the entire file. Reading entire files is often wasteful and slow, especially for large files (i.e. more than a few hundred lines). So you should use this option sparingly. Reading the entire file is not allowed in most cases. You are only allowed to read the entire file if it has been edited or manually attached to the conversation by the user.",
)
def read_file(
        target_file: str,
        start_line_one_indexed: int,
        end_line_one_indexed_inclusive: int,
        should_read_entire_file: bool = False,
        explanation: str | None = None
) -> str:
    """Read the contents of a file. the output of this tool call will be the 1-indexed file contents from start_line_one_indexed to end_line_one_indexed_inclusive, together with a summary of the lines outside start_line_one_indexed and end_line_one_indexed_inclusive. Note that this call can view at most 250 lines at a time. When using this tool to gather information, it's your responsibility to ensure you have the COMPLETE context. Specifically, each time you call this command you should: 1) Assess if the contents you viewed are sufficient to proceed with your task. 2) Take note of where there are lines not shown. 3) If the file contents you have viewed are insufficient, and you suspect they may be in lines not shown, proactively call the tool again to view those lines. 4) When in doubt, call this tool again to gather more information. Remember that partial file views may miss critical dependencies, imports, or functionality. In some cases, if reading a range of lines is not enough, you may choose to read the entire file. Reading entire files is often wasteful and slow, especially for large files (i.e. more than a few hundred lines). So you should use this option sparingly. Reading the entire file is not allowed in most cases. You are only allowed to read the entire file if it has been edited or manually attached to the conversation by the user.

    :param target_file: The path of the file to read. You can use either a relative path in the workspace or an absolute path. If an absolute path is provided, it will be preserved as is.
    :param start_line_one_indexed: The one-indexed line number to start reading from (inclusive).
    :param end_line_one_indexed_inclusive: The one-indexed line number to end reading at (inclusive).
    :param should_read_entire_file: Whether to read the entire file. Defaults to false.
    :param explanation: One sentence explanation as to why this tool is being used, and how it contributes to the goal.
    """
    if explanation:
        print(f"[{constants.AI_AGENT_NAME}] {explanation}", flush=True)

    if not os.path.exists(target_file):
        return f"File {target_file} does not exist."

    try:
        with open(target_file, "r", encoding="utf-8") as f:
            lines = f.readlines()

        total_lines = len(lines)

        # Validate inputs
        if start_line_one_indexed < 1 or end_line_one_indexed_inclusive > total_lines:
            return f"Invalid line range. File has {total_lines} lines."

        if end_line_one_indexed_inclusive < start_line_one_indexed:
            return "End line must be greater than or equal to start line."

        if end_line_one_indexed_inclusive - start_line_one_indexed + 1 > 250:
            return "You can only request up to 250 lines at a time."

        if should_read_entire_file:
            return "".join(lines)

        selected_lines = lines[start_line_one_indexed - 1:end_line_one_indexed_inclusive]
        output = "".join(selected_lines)

        # Add context summary
        before_summary = f"...{len(lines[:start_line_one_indexed - 1])} lines above..." if start_line_one_indexed > 1 else ""
        after_summary = f"...{len(lines[end_line_one_indexed_inclusive:])} lines below..." if end_line_one_indexed_inclusive < total_lines else ""

        if before_summary or after_summary:
            output = f"{before_summary}\n{output}\n{after_summary}"

        print(f"[{constants.AI_AGENT_NAME}]: read_file {target_file} : ({start_line_one_indexed}-{end_line_one_indexed_inclusive})")

        return output

    except Exception as e:
        return f"[Error] Failed to read file: {str(e)}"


@tool(
    name_or_callable="run_terminal_cmd",
    description="PROPOSE a command to run on behalf of the user. If you have this tool, note that you DO have the ability to run commands directly on the USER's system. Note that the user will have to approve the command before it is executed. The user may reject it if it is not to their liking, or may modify the command before approving it. If they do change it, take those changes into account. The actual command will NOT execute until the user approves it. The user may not approve it immediately. Do NOT assume the command has started running. If the step is WAITING for user approval, it has NOT started running. In using these tools, adhere to the following guidelines: 1. Based on the contents of the conversation, you will be told if you are in the same shell as a previous step or a different shell. 2. If in a new shell, you should `cd` to the appropriate directory and do necessary setup in addition to running the command. 3. If in the same shell, the state will persist (eg. if you cd in one step, that cwd is persisted next time you invoke this tool). 4. For ANY commands that would use a pager or require user interaction, you should append ` | cat` to the command (or whatever is appropriate). Otherwise, the command will break. You MUST do this for: git, less, head, tail, more, etc. 5. For commands that are long running/expected to run indefinitely until interruption, please run them in the background. To run jobs in the background, set `is_background` to true rather than changing the details of the command. 6. Dont include any newlines in the command.",
)
def run_terminal_cmd(
        command: str,
        is_background: bool,
        explanation: str | None = None,
):
    """PROPOSE a command to run on behalf of the user. If you have this tool, note that you DO have the ability to run commands directly on the USER's system. Note that the user will have to approve the command before it is executed. The user may reject it if it is not to their liking, or may modify the command before approving it. If they do change it, take those changes into account. The actual command will NOT execute until the user approves it. The user may not approve it immediately. Do NOT assume the command has started running. If the step is WAITING for user approval, it has NOT started running. In using these tools, adhere to the following guidelines: 1. Based on the contents of the conversation, you will be told if you are in the same shell as a previous step or a different shell. 2. If in a new shell, you should `cd` to the appropriate directory and do necessary setup in addition to running the command. 3. If in the same shell, the state will persist (eg. if you cd in one step, that cwd is persisted next time you invoke this tool). 4. For ANY commands that would use a pager or require user interaction, you should append ` | cat` to the command (or whatever is appropriate). Otherwise, the command will break. You MUST do this for: git, less, head, tail, more, etc. 5. For commands that are long running/expected to run indefinitely until interruption, please run them in the background. To run jobs in the background, set `is_background` to true rather than changing the details of the command. 6. Dont include any newlines in the command.

    :param command: The terminal command to execute
    :param is_background: Whether the command should be run in the background
    :param explanation: One sentence explanation as to why this command needs to be run and how it contributes to the goal.
    :return:
    """
    # Rule 4: Prevent pagers from blocking by appending `| cat`
    if any(cmd in command.lower() for cmd in ["git", "less", "head", "tail", "more"]):
        command += " | cat"

    # Output explanation if provided
    if explanation:
        print(f"[{constants.AI_AGENT_NAME}] {explanation}", flush=True)

    # Display proposed command
    print(f"[Proposed Command]: {command}", flush=True)
    # Auto-detect Python-related commands
    python_commands = {"python", "pip", "uv", "gunicorn", "flask"}
    first_token = command.strip().split()[0].lower()

    venv_path = os.path.join(os.getcwd(), ".venv")
    activate_script = ""

    if first_token in python_commands and os.path.isdir(venv_path):
        # Determine activation script based on OS
        if os.name == 'posix':
            activate_script = os.path.join(venv_path, "bin", "activate")
            activate_cmd = f"source {activate_script}"
        else:
            activate_script = os.path.join(venv_path, "Scripts", "activate")
            activate_cmd = activate_script

        print(f"[{constants.AI_AGENT_NAME}] Activating virtual environment at {venv_path}...", flush=True)

        # Wrap original command with activation
        if os.name == 'posix':
            command = f"source {activate_script} && {command}"
        else:
            command = f'{activate_script} && {command}'

    # Run command
    try:
        # Split command safely
        if os.name == 'posix':
            args = ['bash', '-c', command]
        else:
            args = ['cmd', '/c', command]

        # Background execution handling
        if is_background:
            # Run in background using nohup (Unix-like) or start-process (Windows)
            with open("background_output.log", "a") as f:
                process = subprocess.Popen(args, stdout=f, stderr=f)
            return f"Command started in background (PID: {process.pid})"
        else:
            # Run synchronously and capture output
            result = subprocess.run(args, capture_output=True, text=True, check=True, shell=True)
            print(f"[Command Output] {result.stdout}{result.stderr}", flush=True)
            return result.stdout + result.stderr

    except Exception as e:
        return f"[Error] Failed to execute command: {str(e)}"


@tool(
    name_or_callable="list_dir",
    description="List the contents of a directory. The quick tool to use for discovery, before using more targeted tools like semantic search or file reading. Useful to try to understand the file structure before diving deeper into specific files. Can be used to explore the codebase."
)
def list_dir(
        relative_workspace_path: str,
        explanation: str | None = None,
):
    """List the contents of a directory. The quick tool to use for discovery, before using more targeted tools like semantic search or file reading. Useful to try to understand the file structure before diving deeper into specific files. Can be used to explore the codebase.

    :param relative_workspace_path: Path to list contents of, relative to the workspace root.
    :param explanation: One sentence explanation as to why this tool is being used, and how it contributes to the goal.
    :return:
    """
    if explanation:
        print(f"[{constants.AI_AGENT_NAME}] {explanation}", flush=True)

    # Construct full path
    full_path = os.path.join(os.getcwd(), relative_workspace_path)

    if not os.path.exists(full_path):
        return f"Path '{relative_workspace_path}' does not exist."

    if not os.path.isdir(full_path):
        return f"'{relative_workspace_path}' is not a directory."

    try:
        # Get directory contents
        entries = os.listdir(full_path)

        if not entries:
            return f"The directory '{relative_workspace_path}' is empty."

        # Separate files and directories
        dirs = [entry for entry in entries if os.path.isdir(os.path.join(full_path, entry))]
        files = [entry for entry in entries if os.path.isfile(os.path.join(full_path, entry))]

        output = []

        if dirs:
            output.append("Directories:")
            output.extend([f"  - {d}/ (dir)" for d in dirs])

        if files:
            output.append("Files:")
            output.extend([f"  - {f}" for f in files])

        return "\n".join(output)

    except Exception as e:
        return f"[Error] Failed to list directory: {str(e)}"


@tool(
    name_or_callable="grep_search",
    description="Fast text-based regex search that finds exact pattern matches within files or directories, utilizing the ripgrep command for efficient searching. Results will be formatted in the style of ripgrep and can be configured to include line numbers and content. To avoid overwhelming output, the results are capped at 50 matches. Use the include or exclude patterns to filter the search scope by file type or specific paths. This is best for finding exact text matches or regex patterns. More precise than semantic search for finding specific strings or patterns. This is preferred over semantic search when we know the exact symbol/function name/etc. to search in some set of directories/file types.",
)
def grep_search(
        query: str,
        include_pattern: str | None = None,
        exclude_pattern: str | None = None,
        case_sensitive: bool | None = None,
        explanation: str | None = None,
):
    """"Fast text-based regex search that finds exact pattern matches within files or directories, utilizing the ripgrep command for efficient searching. Results will be formatted in the style of ripgrep and can be configured to include line numbers and content. To avoid overwhelming output, the results are capped at 50 matches. Use the include or exclude patterns to filter the search scope by file type or specific paths. This is best for finding exact text matches or regex patterns. More precise than semantic search for finding specific strings or patterns. This is preferred over semantic search when we know the exact symbol/function name/etc. to search in some set of directories/file types."

    :param query: The regex pattern to search for
    :param include_pattern: Glob pattern for files to include (e.g. '*.ts' for TypeScript files)
    :param exclude_pattern: Glob pattern for files to exclude
    :param case_sensitive: Whether the search should be case sensitive
    :param explanation: One sentence explanation as to why this tool is being used, and how it contributes to the goal.
    :return:
    """
    if explanation:
        print(f"[{constants.AI_AGENT_NAME}] {explanation}", flush=True)

        # Build base command
    command = ["rg", "--color=never"]

    # Add case sensitivity flag
    if case_sensitive is False:
        command.append("--ignore-case")
    elif case_sensitive is True:
        command.append("--case-sensitive")

    # Add include/exclude patterns
    if include_pattern:
        command.extend(["--glob", include_pattern])
    if exclude_pattern:
        command.extend(["--glob", f"!{exclude_pattern}"])

    # Always show line numbers and limit to 50 matches
    command.extend(["--line-number", "--max-count=1", query, "."])

    try:
        result = subprocess.run(command, capture_output=True, text=True, check=False)
        output = result.stdout.strip()

        if not output:
            return "No matches found."

        # Limit to 50 matches
        matches = output.split('\n')
        if len(matches) > 50:
            output = "\n".join(matches[:50]) + "\n... (limited to 50 matches)"

        return output

    except FileNotFoundError:
        return "[Error] ripgrep (rg) is not installed. Please install it first."
    except Exception as e:
        return f"[Error] Failed to perform grep search: {str(e)}"


@tool(
    name_or_callable="edit_file",
    description="Use this tool to propose an edit to an existing file. This will be read by a less intelligent model, which will quickly apply the edit. You should make it clear what the edit is, while also minimizing the unchanged code you write. When writing the edit, you should specify each edit in sequence, with the special comment `// ... existing code ...` to represent unchanged code in between edited lines. For example: ``` // ... existing code ... FIRST_EDIT // ... existing code ... SECOND_EDIT // ... existing code ... THIRD_EDIT // ... existing code ... ``` You should still bias towards repeating as few lines of the original file as possible to convey the change. But, each edit should contain sufficient context of unchanged lines around the code you're editing to resolve ambiguity. DO NOT omit spans of pre-existing code (or comments) without using the `// ... existing code ...` comment to indicate its absence. If you omit the existing code comment, the model may inadvertently delete these lines. Make sure it is clear what the edit should be, and where it should be applied. You should specify the following arguments before the others: [target_file]"
)
def edit_file(
        target_file: str,
        explanation: str,
        code_edited: str,
        code_edit: str,
):
    """Use this tool to propose an edit to an existing file. This will be read by a less intelligent model, which will quickly apply the edit. You should make it clear what the edit is, while also minimizing the unchanged code you write. When writing the edit, you should specify each edit in sequence, with the special comment `// ... existing code ...` to represent unchanged code in between edited lines. For example: ``` // ... existing code ... FIRST_EDIT // ... existing code ... SECOND_EDIT // ... existing code ... THIRD_EDIT // ... existing code ... ``` You should still bias towards repeating as few lines of the original file as possible to convey the change. But, each edit should contain sufficient context of unchanged lines around the code you're editing to resolve ambiguity. DO NOT omit spans of pre-existing code (or comments) without using the `// ... existing code ...` comment to indicate its absence. If you omit the existing code comment, the model may inadvertently delete these lines. Make sure it is clear what the edit should be, and where it should be applied. You should specify the following arguments before the others: [target_file]

    :param target_file: The target file to modify. Always specify the target file as the first argument. You can use either a relative path in the workspace or an absolute path. If an absolute path is provided, it will be preserved as is.
    :param explanation: A single sentence instruction describing what you are going to do for the sketched edit. This is used to assist the less intelligent model in applying the edit. Please use the first person to describe what you are going to do. Dont repeat what you have said previously in normal messages. And use it to disambiguate uncertainty in the edit.
    :param code_edited: The exact block of code to find and replace.
    :param code_edit: Specify ONLY the precise lines of code that you wish to edit. **NEVER specify or write out unchanged code**. Instead, represent all unchanged code using the comment of the language you're editing in - example: `// ... existing code ...`
    :return:
    """

    print(f"[{constants.AI_AGENT_NAME}] {explanation}", flush=True)

    if not os.path.exists(target_file):
        return f"File {target_file} does not exist."

    try:
        # Read original file content
        with open(target_file, "r", encoding="utf-8") as f:
            original_content = f.read()

        # Check if old_code exists exactly
        if code_edited not in original_content:
            return f"[Warning] Could not find exact match for old_code. Edit was not applied."

        edit_lines = []
        for line in code_edit.split('\n'):
            strip_line = line.strip()
            if strip_line.startswith('//') or strip_line.startswith('#'):
                continue
            edit_lines.append(line)

        # Replace old_code with new_code
        updated_content = original_content.replace(code_edited, '\n'.join(edit_lines))

        # Write back to file
        with open(target_file, "w", encoding="utf-8") as f:
            f.write(updated_content)

        if codebase:
            codebase.update_file_documents(target_file)

        return f"Successfully replaced code in {target_file}"

    except Exception as e:
        return f"[Error] Failed to edit file: {str(e)}"


# @tool(
#     name_or_callable="edit_file",
#     description="Use this tool to propose an edit to an existing file. This will be read by a less intelligent model, which will quickly apply the edit. You should make it clear what the edit is, while also minimizing the unchanged code you write. When writing the edit, you should specify each edit in sequence, with the special comment `// ... existing code ...` to represent unchanged code in between edited lines. For example: ``` // ... existing code ... FIRST_EDIT // ... existing code ... SECOND_EDIT // ... existing code ... THIRD_EDIT // ... existing code ... ``` You should still bias towards repeating as few lines of the original file as possible to convey the change. But, each edit should contain sufficient context of unchanged lines around the code you're editing to resolve ambiguity. DO NOT omit spans of pre-existing code (or comments) without using the `// ... existing code ...` comment to indicate its absence. If you omit the existing code comment, the model may inadvertently delete these lines. Make sure it is clear what the edit should be, and where it should be applied. You should specify the following arguments before the others: [target_file]"
# )
# def edit_file(
#         target_file: str,
#         instructions: str,
#         code_edit: str,
# ):
#     """Use this tool to propose an edit to an existing file. This will be read by a less intelligent model, which will quickly apply the edit. You should make it clear what the edit is, while also minimizing the unchanged code you write. When writing the edit, you should specify each edit in sequence, with the special comment `// ... existing code ...` to represent unchanged code in between edited lines. For example: ``` // ... existing code ... FIRST_EDIT // ... existing code ... SECOND_EDIT // ... existing code ... THIRD_EDIT // ... existing code ... ``` You should still bias towards repeating as few lines of the original file as possible to convey the change. But, each edit should contain sufficient context of unchanged lines around the code you're editing to resolve ambiguity. DO NOT omit spans of pre-existing code (or comments) without using the `// ... existing code ...` comment to indicate its absence. If you omit the existing code comment, the model may inadvertently delete these lines. Make sure it is clear what the edit should be, and where it should be applied. You should specify the following arguments before the others: [target_file]
#
#     :param target_file: The target file to modify. Always specify the target file as the first argument. You can use either a relative path in the workspace or an absolute path. If an absolute path is provided, it will be preserved as is.
#     :param instructions: A single sentence instruction describing what you are going to do for the sketched edit. This is used to assist the less intelligent model in applying the edit. Please use the first person to describe what you are going to do. Dont repeat what you have said previously in normal messages. And use it to disambiguate uncertainty in the edit.
#     :param code_edit: Specify ONLY the precise lines of code that you wish to edit. **NEVER specify or write out unchanged code**. Instead, represent all unchanged code using the comment of the language you're editing in - example: `// ... existing code ...`
#     :return:
#     """
#     print(f"[{constants.AI_AGENT_NAME}] {instructions}")
#
#     if not os.path.exists(target_file):
#         return f"File {target_file} does not exist."
#
#     try:
#         # Read original file content
#         with open(target_file, "r", encoding="utf-8") as f:
#             original_lines = f.readlines()
#
#         # Parse the code_edit content into blocks
#         edit_blocks = []
#         current_block = {"before": [], "replace_with": []}
#         in_existing = True  # We start outside any edit block
#
#         for line in code_edit.strip().split("\n"):
#             stripped_line = line.strip()
#             if stripped_line == "// ... existing code ...":
#                 if in_existing:
#                     # Start new edit block
#                     current_block["after"] = []
#                     edit_blocks.append(current_block)
#                     current_block = {"before": [], "replace_with": []}
#                 else:
#                     # End current edit block, prepare next one
#                     current_block["after"] = []
#                     edit_blocks.append(current_block)
#                     current_block = {"before": [], "replace_with": []}
#                 in_existing = True
#             else:
#                 current_block["replace_with"].append(line)
#                 in_existing = False
#
#         # Append the last block if it has content
#         if current_block["replace_with"]:
#             current_block["after"] = []
#             edit_blocks.append(current_block)
#
#         # Apply edits to the original lines
#         new_content = []
#         matched = False
#         i = 0  # Current index in original_lines
#         n = len(original_lines)
#
#         for block in edit_blocks:
#             before_context = block["before"]
#             replace_with = block["replace_with"]
#
#             # Try to find matching block in original lines
#             found_index = -1
#             for start in range(n - len(before_context) + 1):
#                 match = True
#                 for j, ctx_line in enumerate(before_context):
#                     if start + j >= n or original_lines[start + j].rstrip('\n') != ctx_line.rstrip('\n'):
#                         match = False
#                         break
#                 if match:
#                     found_index = start + len(before_context)
#                     break
#
#             if found_index != -1:
#                 # Replace the next lines after match
#                 new_content.extend(original_lines[i:start])
#                 new_content.extend([line + "\n" for line in replace_with])
#                 i = found_index
#                 matched = True
#             else:
#                 # Fallback: append replacement at end (context not found)
#                 new_content.extend([line + "\n" for line in replace_with])
#                 matched = False
#
#         # Append remaining lines
#         new_content.extend(original_lines[i:])
#
#         if not matched:
#             return "[Warning] Could not find exact match for context. Appended edit at end of file."
#
#         # Write back to file
#         with open(target_file, "w", encoding="utf-8") as f:
#             f.writelines(new_content)
#
#         return f"Successfully applied edit to {target_file}"
#
#     except Exception as e:
#         return f"[Error] Failed to edit file: {str(e)}"

@tool(
    name_or_callable="file_search",
    description="Fast file search based on fuzzy matching against file path. Use if you know part of the file path but don't know where it's located exactly. Response will be capped to 10 results. Make your query more specific if need to filter results further.",
)
def file_search(
        query: str,
        explanation: str | None = None,
):
    """Fast file search based on fuzzy matching against file path. Use if you know part of the file path but don't know where it's located exactly. Response will be capped to 10 results. Make your query more specific if need to filter results further.

    :param query: Fuzzy filename to search for
    :param explanation: One sentence explanation as to why this tool is being used, and how it contributes to the goal.
    :return:
    """
    if explanation:
        print(f"[{constants.AI_AGENT_NAME}] {explanation}", flush=True)

        # Normalize query for case-insensitive matching
    query_lower = query.lower()
    matches: list[str] = []

    # Walk through all files and directories
    for root, dirs, files in os.walk(os.getcwd()):
        for name in files + dirs:
            full_path = os.path.join(root, name)

            # Skip hidden/system files/folders (optional)
            if any(part.startswith('.') or part == "__pycache__" for part in full_path.split(os.sep)):
                continue

            # Fuzzy match logic: check if query appears in the path (case-insensitive)
            if query_lower in full_path.lower():
                matches.append(full_path)

    # Limit to 10 results
    matches = matches[:10]

    if not matches:
        return f"No files found matching '{query}'."

    # Format output
    result_lines = ["Matched files/directories:"]
    for path in matches:
        result_lines.append(f"  - {path}")

    return "\n".join(result_lines)


@tool(
    name_or_callable="create_file",
    description="Creates a new file at the specified path with optional initial content. The operation will fail gracefully if: - The file already exists - The operation is rejected for security reasons - The file cannot be created due to permission or system issues",
)
def create_file(
        target_file: str,
        initial_content: str = '',
        explanation: str | None = None,
):
    """Creates a new file at the specified path with optional initial content. The operation will fail gracefully if: - The file already exists - The operation is rejected for security reasons - The file cannot be created due to permission or system issues

    :param target_file: The path of the file to create, relative to the workspace root.
    :param initial_content: Optional initial content to write into the newly created file.
    :param explanation: One sentence explanation as to why this tool is being used, and how it contributes to the goal.
    :return:
    """
    if explanation:
        print(f"[{constants.AI_AGENT_NAME}] {explanation}", flush=True)

    full_path = os.path.abspath(target_file)

    # Check if file already exists
    if os.path.exists(full_path):
        return f"File '{target_file}' already exists."

    try:
        # Ensure directory exists
        os.makedirs(os.path.dirname(full_path), exist_ok=True)

        # Create file and optionally write content
        with open(full_path, "w", encoding="utf-8") as f:
            f.write(initial_content)

        if codebase:
            try:
                codebase.add_file_documents(full_path)
            except Exception as e:
                print(f'[{constants.AI_AGENT_NAME}] Codebase add file index error.')

        return f"Successfully created file: {target_file}"
    except PermissionError:
        return f"[Error] Permission denied when trying to create '{target_file}'."
    except Exception as e:
        return f"[Error] Failed to create file: {str(e)}"


@tool(
    name_or_callable="delete_file",
    description="Deletes a file at the specified path. The operation will fail gracefully if: - The file doesn't exist - The operation is rejected for security reasons - The file cannot be deleted",
)
def delete_file(
        target_file: str,
        explanation: str | None = None,
):
    """Deletes a file at the specified path. The operation will fail gracefully if: - The file doesn't exist - The operation is rejected for security reasons - The file cannot be deleted

    :param target_file: The path of the file to delete, relative to the workspace root.
    :param explanation: One sentence explanation as to why this tool is being used, and how it contributes to the goal.
    :return:
    """
    if explanation:
        print(f"[{constants.AI_AGENT_NAME}] {explanation}", flush=True)

    full_path = os.path.abspath(target_file)

    # Check if file exists
    if not os.path.exists(full_path):
        return f"File '{target_file}' does not exist."

    # Ensure it's a file (not a directory)
    if not os.path.isfile(full_path):
        return f"'{target_file}' is not a file."

    try:
        os.remove(full_path)
        if codebase:
            codebase.remove_file_documents(full_path)
        return f"Successfully deleted file: {target_file}"
    except PermissionError:
        return f"[Error] Permission denied when trying to delete '{target_file}'."
    except Exception as e:
        return f"[Error] Failed to delete file: {str(e)}"


@tool(
    name_or_callable="reapply",
    description="Calls a smarter model to apply the last edit to the specified file. Use this tool immediately after the result of an edit_file tool call ONLY IF the diff is not what you expected, indicating the model applying the changes was not smart enough to follow your instructions.",
)
def reapply(
        target_file: str,
):
    """Calls a smarter model to apply the last edit to the specified file. Use this tool immediately after the result of an edit_file tool call ONLY IF the diff is not what you expected, indicating the model applying the changes was not smart enough to follow your instructions.

    :param target_file: The relative path to the file to reapply the last edit to. You can use either a relative path in the workspace or an absolute path. If an absolute path is provided, it will be preserved as is.
    :return:
    """
    # todo: implements
    raise NotImplemented()


@tool(
    name_or_callable="web_search",
    description="Search the web for real-time information about any topic. Use this tool when you need up-to-date information that might not be available in your training data, or when you need to verify current facts. The search results will include relevant snippets and URLs from web pages. This is particularly useful for questions about current events, technology updates, or any topic that requires recent information."
)
def web_search(
        search_term: str,
        explanation: str | None = None,
):
    """Search the web for real-time information about any topic. Use this tool when you need up-to-date information that might not be available in your training data, or when you need to verify current facts. The search results will include relevant snippets and URLs from web pages. This is particularly useful for questions about current events, technology updates, or any topic that requires recent information.

    :param search_term: The search term to look up on the web. Be specific and include relevant keywords for better results. For technical queries, include version numbers or dates if relevant.
    :param explanation: One sentence explanation as to why this tool is being used, and how it contributes to the goal.
    :return:
    """
    # todo: implements
    raise NotImplemented()


@tool(
    name_or_callable="diff_history",
    description="Retrieve the history of recent changes made to files in the workspace. This tool helps understand what modifications were made recently, providing information about which files were changed, when they were changed, and how many lines were added or removed. Use this tool when you need context about recent modifications to the codebase."
)
def diff_history(
        explanation: str | None = None,
):
    """Retrieve the history of recent changes made to files in the workspace. This tool helps understand what modifications were made recently, providing information about which files were changed, when they were changed, and how many lines were added or removed. Use this tool when you need context about recent modifications to the codebase.

    :param explanation: One sentence explanation as to why this tool is being used, and how it contributes to the goal.
    :return:
    """
    # todo: implements
    raise NotImplemented()


TOOLS = [
    codebase_search,
    read_file,
    run_terminal_cmd,
    list_dir,
    grep_search,
    edit_file,
    file_search,
    create_file,
    delete_file,
]



import os
from typing import Type, Callable

from gitignore_parser import parse_gitignore, parse_gitignore_str
from langchain_community.document_loaders import TextLoader
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_core.vectorstores import InMemoryVectorStore, VectorStore
from langchain_text_splitters import RecursiveCharacterTextSplitter, Language

from elpis import i18n
from elpis.factories import model_factory


class CodebaseIndexer:
    """A simple codebase and index"""

    def __init__(self, workspace: str = os.getcwd(), embeddings: Embeddings = None,
                 vector_store_cls: Type[VectorStore] = InMemoryVectorStore,
                 text_chunk_size: int = 1000, text_chunk_overlap: int = 200,
                 code_chunk_size: int = 1000, code_chunk_overlap: int = 0):
        self._workspace = os.path.abspath(workspace)
        if embeddings:
            self._embeddings = embeddings
        else:
            self._embeddings = model_factory.new_model(
                os.getenv('EMBEDDING_MODEL_KEY_PREFIX')
            )
        self._vector_store_cls = vector_store_cls
        self._vectorstore = None
        self._lang = i18n.select_lang(os.getenv('LANG', default='zh'))
        self._text_loader = TextLoader(
            file_path='',
            encoding='utf-8',
            autodetect_encoding=True,
        )
        self._file_suffix_code_mapping = {
            'py': 'python',
            'pyi': 'python',
            'cpp': 'cpp',
            'go': 'go',
            'java': 'java',
            'kt': 'kotlin',
            'kts': 'kotlin',
            'js': 'js',
            'jsx': 'js',
            'mjs': 'js',
            'cjs': 'js',
            'ts': 'ts',
            'tsx': 'ts',
            'php': 'php',
            'proto': 'proto',
            'rst': 'rst',
            'rb': 'ruby',
            'rs': 'rust',
            'scala': 'scala',
            'swift': 'swift',
            'md': 'markdown',
            'markdown': 'markdown',
            'tex': 'latex',
            'latex': 'latex',
            'html': 'html',
            'htm': 'html',
            'sol': 'sol',
            'cs': 'csharp',
            'cobol': 'cobol',
            'c': 'c',
            'h': 'c',  # C header files
            'hh': 'cpp',  # C++ header files
            'hpp': 'cpp',  # C++ header files
            'lua': 'lua',
            'pl': 'perl',
            'pm': 'perl',
            'hs': 'haskell',
        }
        self._splitters = {}
        self._text_chunk_size = text_chunk_size
        self._text_chunk_overlap = text_chunk_overlap
        self._code_chunk_size = code_chunk_size
        self._code_chunk_overlap = code_chunk_overlap

    def _load_file_data(self, file_path: str) -> list[Document]:
        try:
            return TextLoader(file_path=file_path, encoding='utf-8', autodetect_encoding=True).load()
        except Exception:
            return []

    def _is_code_file(self, file_suffix: str):
        return file_suffix in self._file_suffix_code_mapping

    def get_current_gitignore_matchers(self) -> Callable[[str], bool] | None:
        default_gitignore_matchers = parse_gitignore_str(""".mypy_cache/
/.coverage
/.coverage.*
/.nox/
/.python-version
/.pytype/
/dist/
/docs/_build/
/src/*.egg-info/
__pycache__/
.idea/
.vscode/
build/
dist/
logs/
venv/
.venv/
node_modules/
""", base_dir=self._workspace)
        gitignore_path = os.path.join(self._workspace, '.gitignore')
        if os.path.exists(gitignore_path):
            return lambda f: default_gitignore_matchers(f) or parse_gitignore(gitignore_path, base_dir=self._workspace)(f)
        return default_gitignore_matchers

    def index_codebase(self):
        print(f'[codebase]: {self._lang.CODEBASE_START_INDEX} {self._workspace}', flush=True)

        gitignore_matchers = self.get_current_gitignore_matchers()

        documents = []

        for root, dirs, files in os.walk(self._workspace):
            # 构建当前目录相对于工作空间的路径
            relative_root = os.path.relpath(root, self._workspace)

            # 过滤掉被 .gitignore 忽略的目录（提前剪枝）
            filtered_dirs = []
            for d in dirs:
                dir_path = os.path.join(relative_root, d)
                # ignore .gitignore
                if gitignore_matchers and gitignore_matchers(dir_path):
                    continue
                filtered_dirs.append(d)
            dirs[:] = filtered_dirs  # 替换，实现剪枝

            for file in files:
                file_path = os.path.join(root, file)
                documents += self.add_file_documents(file_path, gitignore_matchers)

        print(self._lang.CODEBASE_FINISH_INDEX.format(len(documents)))

    def search_codebase(self, query: str, k: int = 5) -> list[Document]:
        if not self._vectorstore:
            return []

        return self._vectorstore.similarity_search(query, k=k)

    def add_file_documents(self, file_path: str,
                           gitignore_matchers: Callable[[str], bool] | None = None):

        file_suffix = file_path.split('.')[-1]
        relative_path = os.path.relpath(file_path, self._workspace)

        # ignore .gitignore
        if gitignore_matchers and gitignore_matchers(relative_path):
            return []

        current_mtime = os.path.getmtime(file_path)

        # load file as documents
        current_documents = self._load_file_data(file_path)

        if not current_documents:
            return []

        splitter_key = self._file_suffix_code_mapping[file_suffix] if file_suffix in self._file_suffix_code_mapping else 'text'
        splitter = self._splitters.get(file_suffix)
        if splitter is None:
            if splitter_key == 'text':
                splitter = RecursiveCharacterTextSplitter(
                    chunk_size=self._text_chunk_size,
                    chunk_overlap=self._text_chunk_overlap,
                    length_function=len,
                    is_separator_regex=False,
                )
            else:
                splitter = RecursiveCharacterTextSplitter.from_language(
                    Language(splitter_key),
                    chunk_size=self._code_chunk_size,
                    chunk_overlap=self._code_chunk_overlap,
                )
            self._splitters[splitter_key] = splitter

        current_documents = splitter.split_documents(current_documents)

        # 更新 mtime 记录
        if self._embeddings and not self._vectorstore:
            self._vectorstore = self._vector_store_cls.from_documents(current_documents, self._embeddings)
        else:
            self._vectorstore.add_documents(current_documents)
        return current_documents

    def remove_file_documents(self, file_path: str):
        if not self._vectorstore:
            return
        file_path = os.path.abspath(file_path)
        if hasattr(self._vectorstore, 'delete'):
            self._vectorstore.delete(filter={'source': file_path})

    def update_file_documents(self, file_path: str):
        self.remove_file_documents(file_path)
        self.add_file_documents(file_path)
